// TODO: remove OSC here as well?
import processing.core.PApplet;
import processing.core.PFont;
import processing.core.PImage;
import codeanticode.gsvideo.*;
import oscP5.*;
import netP5.*;


/**
 *
 */
private static final long serialVersionUID = 1L;

/*for the full screen
*
* public void init() {
*    frame.removeNotify();
*     frame.setUndecorated(true);
*     frame.addNotify();
*     super.init();
* }
* public static void main(String[] args) {
*     PApplet.main(new String[] { "toonLoop.Main" });
* }
*/

/**
 * ToonLoop :: Realtime stop motion performance tool.
 *
 * @author Alexandre Quessy <alexandre@quessy.net>
 * @license GNU Public License version 2
 *
 *          In the left window, you can see what is seen by the live camera.
 *          In the right window, it is the result of the stop motion loop.
 *
 *          Usage : - Press space bar to grab a frame. - Press DELETE or
 *          BACKSPACE to delete the last frame. - Press 'r' to reset and
 *          start the current sequence. (an remove all its frames) - Press
 *          's' to save the current sequence as a QuickTime movie. - Press
 *          'p' to open the Quicktime video camera settings dialog. (if
 *          available) - Press 'a' to toggle on/off the auto recording. (it
 *          records one frame on every frame) - Press 'q' to quit (disabled)
 *          - Press UP to increase frame rate - Press DOWN to decrease frame
 *          rate - Press numbers from 0 to 9 to switch to an other sequence.
 *
 *          Installation : - Run the application in the application.macosx/
 *          or application.windows/ folder. It should work. If not, you need
 *          to compile the sketch as indicated in "Using with Processing".
 *
 *          Using with Processing : - You need Processing
 *          (http://processing.org). - Put the toonloop folder in
 *          ~/Documents/Processing/ (where "~" represents your home
 *          directory, such as "/Users/alexandre/") - Change
 *          Processing::Preferences::Set maximum available memory to... to a
 *          higher value, such as 500 Mb at least. - Set LOOP_MAX_NUM_FRAME
 *          to the desired number of frames. - Run the sketch. - Quit the
 *          applet to stop. - Close the window and start over to create a
 *          new movie. - Look in the toonloop directory for the saved movie
 *          files.
 *
 * @version 0.13 2008-12-15 PureData patch toonloop.midi.osc.control.pd to
 *          latest version. Use it to control ToonLoop with a MIDI keyboard.
 *          Now saving in a thread. (no impact on rendering speed)
 *          Displaying sequence number in red at the bottom.
 * @version 0.12 2008-12-05 Added OSC listeners: /toon/grab
 * @version 0.11 2008-12-02 Now with a bigger window Added UP and DOWN
 *          arrows to increase/decrease frame rate. Now saving in a thread.
 * @version 0.10 2008-11-09 Each sequence has its own frame rate Adding OSC
 *          sent messages : /frame, /sequence, /editframe and /framerate
 *          Added ENABLE_QUIT configuration variable.
 * @version 0.9 2008-11-08 Added OSC protocol messages Now multiples
 *          sequences Interfacing with FxLoop, the Ambisonic sound effects
 *          sequencer Pure Data patch
 * @version 0.7 2008-09-27 Improving usability Added features: delete
 *          frames, reset movie, auto record. Configuration with variables.
 *          Frames numbering starts at 1.
 * @license GNU General Public License version 2 Contact the author if you
 *          want a commercial version and license of this tool.
 * @url http://alexandre.quessy.net/?q=toonloop
 */
// ---------------------------- IMPORT LIBRARIES

int OSC_RECEIVE_PORT = 31338;
int OSC_SEND_PORT = 31337;
String OSC_SEND_HOST = "127.0.0.1";
// ---------------- CONFIGURATION -------------------//
int LOOP_MAX_NUM_FRAME = 500; // Max number of frames.
int NUM_SEQUENCES = 10; // Numbers of sequences (bins) from 1 to 10
int LOOP_WIDTH = 320; // width of the image from the camera
int LOOP_HEIGHT = 240; // height of the image from the camera
int FRAME_RATE = 8; // frame rate (at startup)
float WINDOW_SIZE_RATIO = 1.5F; // 2; // How bigger the window is compared
                                // to LOOP_WIDTH * 2
// ------------------------------------------
int SAVED_MESSAGE_DURATION = 30; // frames
float SAVED_MESS_SIZE_RATIO = 0.5F;
int TEXT_FONT_SIZE = 24;
boolean ENABLE_QUIT = false;
/* Choose only one of the followings: */
// int LOOP_CODEC = MovieMaker.H263;
// int LOOP_CODEC = MovieMaker.H264;
// int LOOP_CODEC = MovieMaker.MOTION_JPEG_A;
// int LOOP_CODEC = MovieMaker.MOTION_JPEG_B;
// int LOOP_CODEC = MovieMaker.ANIMATION;
/* Choose only one of the followings: */
// int LOOP_QUALITY = MovieMaker.HIGH;
// int LOOP_QUALITY = MovieMaker.BEST;
// int LOOP_QUALITY = MovieMaker.LOSSLESS;
// -------------------------------------------------//
int currentSeq = 0;
OscP5 osc_server;
NetAddress oscRemoteLocation;
GSCapture cam;
PFont font;
int is_auto_recording = 0;
int isFlashing = 0;
int is_displaying_saved_message = 0; // also a counter
// MovieMaker movieOut;
String saved_file_name;
ToonSequence sequences[] = new ToonSequence[NUM_SEQUENCES];

public void setup() {
    size((int) (LOOP_WIDTH * 2 * WINDOW_SIZE_RATIO),
            (int) (LOOP_HEIGHT * 2 * WINDOW_SIZE_RATIO)); // *WINDOW_SIZE_RATIO
    frameRate(FRAME_RATE);

    // println("Available cameras :");
    // print(Capture.list());

    cam = new GSCapture(this, 640, 480);

    for (int i = 0; i < sequences.length; i++) {
        sequences[i] = new ToonSequence();
    }

    font = loadFont("CourierNewPSMT-24.vlw");
    // OSC stuff
    osc_server = new OscP5(this, 31337);
    oscRemoteLocation = new NetAddress(OSC_SEND_HOST, OSC_SEND_PORT);
    osc_server = new OscP5(this, OSC_RECEIVE_PORT);
    oscSendFrameRate(FRAME_RATE);
    println("Welcome to ToonLoop ! The realtime stop motion performance tool.");
    println(")c( Alexandre Quessy 2008");
    println("http://alexandre.quessy.net");
}

/*
 * changes the remote host
 */
void setOscSendHost(String h) {
    OSC_SEND_HOST = h; // sends the former one to the garbage collector
    oscRemoteLocation = new NetAddress(OSC_SEND_HOST, OSC_SEND_PORT);
    println("Changing OSC send host to " + h);
}

void setOscSendPort(int p) {
    OSC_SEND_PORT = p; // sends the former one to the garbage collector
    oscRemoteLocation = new NetAddress(OSC_SEND_HOST, OSC_SEND_PORT);
    println("Changing OSC send port to " + p);
}

public void draw() {
    background(0);
    if (cam.available() == true) {
        cam.read();
    }
    // Live camera image on the left:
    // with message "AUTO RECORDING" if recording
    noTint(); // alpha 100%
    textFont(font, TEXT_FONT_SIZE); // size in pixels

    if (is_auto_recording == 1) {
        sequences[currentSeq].addFrame();
        fill(255, 0, 0, 255);
        // String warningText = "Auto recording on. Press 'a' to stop.";
        String warningText = "AUTO RECORDING";

        text(warningText, (int) (LOOP_WIDTH / (4 * WINDOW_SIZE_RATIO)),
                (int) (LOOP_HEIGHT * 1.7 * WINDOW_SIZE_RATIO));
    } else {
        fill(255, 255, 255, 255);
        if (isFlashing == 1) {
            isFlashing = 0;
            rect(0, (LOOP_HEIGHT / 2) * WINDOW_SIZE_RATIO, LOOP_WIDTH
                    * WINDOW_SIZE_RATIO, LOOP_HEIGHT * WINDOW_SIZE_RATIO); // *WINDOW_SIZE_RATIO
            tint(255, 191); // alpha 75%
        }
    }
    image(cam, 0, (int) ((LOOP_HEIGHT / 2) * WINDOW_SIZE_RATIO),
            (int) (LOOP_WIDTH * WINDOW_SIZE_RATIO),
            (int) (LOOP_HEIGHT * WINDOW_SIZE_RATIO)); // *WINDOW_SIZE_RATIO
    // image(img, x, y, width, height)

    // Frame number under the images :
    fill(127);
    textFont(font, TEXT_FONT_SIZE); // size in pixels

    int y = (int) (LOOP_HEIGHT * 1.6 * WINDOW_SIZE_RATIO);
    if (sequences[currentSeq].captureFrameNum > 0) {
        text("" + sequences[currentSeq].captureFrameNum,
                (int) ((LOOP_WIDTH / 2) * WINDOW_SIZE_RATIO), y);
        text("" + (sequences[currentSeq].playFrameNum + 1),
                (int) ((LOOP_WIDTH * 3 / 2.0) * WINDOW_SIZE_RATIO), y); // 2*WINDOW_SIZE_RATIO,

    }
    fill(127, 0, 0); // darker
    text("" + (currentSeq + 1), (int) ((LOOP_WIDTH) * WINDOW_SIZE_RATIO), y);// TODO:
                                                                                // check
                                                                                // if
                                                                                // it
                                                                                // is
                                                                                // ok

    fill(255);
    // Image at the current offset in the loop on the right:
    noTint(); // alpha 100%
    if (sequences[currentSeq].captureFrameNum > 0) {
        image(
                sequences[currentSeq].images[sequences[currentSeq].playFrameNum],
                (int) (LOOP_WIDTH * WINDOW_SIZE_RATIO),
                (int) ((LOOP_HEIGHT / 2) * WINDOW_SIZE_RATIO),
                (int) (LOOP_WIDTH * WINDOW_SIZE_RATIO),
                (int) (LOOP_HEIGHT * WINDOW_SIZE_RATIO)); // *WINDOW_SIZE_RATIO
        oscSendFrameNum(sequences[currentSeq].playFrameNum);
        sequences[currentSeq].loopFrame();
    }

    // SAVED message
    textFont(font, (int) (TEXT_FONT_SIZE * SAVED_MESS_SIZE_RATIO)); // size
                                                                    // in
                                                                    // pixels
    if (is_displaying_saved_message > 0) {
        is_displaying_saved_message--; // decrement duration
        int x = (int) ((LOOP_WIDTH / 4) * WINDOW_SIZE_RATIO);
        int the_y = (int) (LOOP_HEIGHT * 1.7 * WINDOW_SIZE_RATIO);// (int)(LOOP_HEIGHT*1.8*WINDOW_SIZE_RATIO);
        fill(255, 0, 0, 255);
        text("Saved to " + saved_file_name + ". (" + LOOP_WIDTH + "x"
                + LOOP_HEIGHT + ")", x, the_y);
    }
}

/**
 * CONTROLS
 *
 * Because of how operating systems handle key repeats, holding down a key
 * may cause multiple calls to keyPressed() (and keyReleased() as well). The
 * rate of repeat is set by the operating system and how each computer is
 * configured.
 */
public void keyPressed() {
    switch (keyCode) {
    case UP:
        increaseFrameRate();
        break;
    case DOWN:
        decreaseFrameRate();
        break;
    case LEFT:
        break;
    case RIGHT:
        break;
    // ALT, CONTROL, SHIFT
    }

    switch (key) {
    case ' ': // ADD FRAME
        if (is_auto_recording == 0) {
            sequences[currentSeq].addFrame();
        }
        break;
    case 'p': // SETTINGS
        cameraPreferences();
        break;
    case 'r': // RESET
        sequences[currentSeq].resetMovie();
        break;
    case 's':
        if (is_auto_recording == 0) {
            saveMovie();
        }
        break;
    case BACKSPACE: // DELETE ONE FRAME
    case DELETE:
        sequences[currentSeq].deleteFrame();
        break;
    case 'a': // AUTO RECORD
        toggleAutoRecording();
        break;
    case ESC: // FULL SCREEN
        // todo (OpenGL)
        break;
    case 'q': // QUIT APPLICATION
        if (ENABLE_QUIT) {
            exit();
        }
        break;
    case '0':
        switchToSequence(0);
        break;
    case '1':
        switchToSequence(1);
        break;
    case '2':
        switchToSequence(2);
        break;
    case '3':
        switchToSequence(3);
        break;
    case '4':
        switchToSequence(4);
        break;
    case '5':
        switchToSequence(5);
        break;
    case '6':
        switchToSequence(6);
        break;
    case '7':
        switchToSequence(7);
        break;
    case '8':
        switchToSequence(8);
        break;
    case '9':
        switchToSequence(9);
        break;
    }
}

// switches to an other sequence
void switchToSequence(int i) {
    if (NUM_SEQUENCES > i && i >= 0) {
        currentSeq = i;
        FRAME_RATE = sequences[currentSeq].getFrameRate();
        frameRate(FRAME_RATE);
        oscSendSeqNum(i);
    }
}

void cameraPreferences() {
    println("Opening the settings window. (if available)");
    // cam.settings();
}

/*
 * incoming osc message are forwarded to the oscEvent method. Implemented
 * messages: /framerate <i> /toon/grab /toon/osc/send/host <s>
 */
void oscEvent(OscMessage mess) {
    // TODO: /toon/frame/rate <i>
    // println("Recv OSC mess:");
    // mess.print();
    if (mess.checkAddrPattern("/toon/framerate/set")) { // "/toon/framerate/set"))
                                                        // {
        if (mess.checkTypetag("i")) {
            changeFrameRate(mess.get(0).intValue());
        }
    } else if (mess.checkAddrPattern("/toon/framerate/increase")) {
        increaseFrameRate();
    } else if (mess.checkAddrPattern("/toon/framerate/decrease")) {
        decreaseFrameRate();
    } else if (mess.checkAddrPattern("/toon/grab")) {
        sequences[currentSeq].addFrame();
    } else if (mess.checkAddrPattern("/toon/save")) {
        saveMovie();
    } else if (mess.checkAddrPattern("/toon/reset")) {
        sequences[currentSeq].resetMovie();
    } else if (mess.checkAddrPattern("/toon/quicktime/preferences")) {
        cameraPreferences();
    } else if (mess.checkAddrPattern("/toon/toggle_auto")) {
        toggleAutoRecording();
    } else if (mess.checkAddrPattern("/toon/osc/send/host")) {
        if (mess.checkTypetag("s")) {
            setOscSendHost(mess.get(0).stringValue());
        }
    } else if (mess.checkAddrPattern("/toon/osc/send/port")) {
        if (mess.checkTypetag("i")) {
            setOscSendPort(mess.get(0).intValue());
        }
    } else if (mess.checkAddrPattern("/toon/sequence")) {
        if (mess.checkTypetag("i")) {
            switchToSequence(mess.get(0).intValue());
        }
    } else if (mess.checkAddrPattern("/toon/delete_frame")) {
        sequences[currentSeq].deleteFrame();
    }
}

void increaseFrameRate() {
    changeFrameRate(sequences[currentSeq].getFrameRate() + 1);
}

void decreaseFrameRate() {
    changeFrameRate(sequences[currentSeq].getFrameRate() - 1);
}

void changeFrameRate(int r) {
    if (r > 0 && r <= 60) {
        FRAME_RATE = r;
        frameRate(FRAME_RATE);
        sequences[currentSeq].setFrameRate(FRAME_RATE);
    }
}

void oscSendFrameNum(int num) {
    OscMessage myMessage = new OscMessage("/frame"); // TODO: change for
                                                        // /fxloop/frame/num
                                                        // <i>
    myMessage.add(num + 1); /* add an int to the osc message */
    osc_server.send(myMessage, oscRemoteLocation);
}

void oscSendEditFrameNum(int num) {
    OscMessage myMessage = new OscMessage("/editframe"); // TODO:
                                                            // /fxloop/frame/edit
                                                            // <i>
    myMessage.add(num + 1); /* add an int to the osc message */
    osc_server.send(myMessage, oscRemoteLocation);
}

void oscSendSeqNum(int num) {
    OscMessage myMessage = new OscMessage("/sequence"); // TODO:
                                                        // /fxloop/sequence
                                                        // <i>
    myMessage.add(num); /* add an int to the osc message */
    osc_server.send(myMessage, oscRemoteLocation);
}

void oscSendFrameRate(int num) {
    OscMessage myMessage = new OscMessage("/framerate"); // TODO
                                                            // /fxloop/frame/rate
                                                            // <i>
    myMessage.add(num); /* add an int to the osc message */
    osc_server.send(myMessage, oscRemoteLocation);
}

/**
 * Starts a thread to save a movie of the current sequence in the
 * background.
 */
void saveMovie() 
{
  SaveThread t = new SaveThread(this, currentSeq);
  t.start();
}

void toggleAutoRecording()
{
    if (is_auto_recording == 0)
    {
      is_auto_recording = 1;
    } else {
      is_auto_recording = 0;
    }
}

/**
 * One sequence (bin)
 */
class ToonSequence
{
  int captureFrameNum = 0; //the next captured frame number ... might wrap around
  int playFrameNum = 0;
  PImage[] images = new PImage[LOOP_MAX_NUM_FRAME];
  int seqFrameRate = FRAME_RATE;

  int getFrameRate() 
  {
    return seqFrameRate;
  }
  void setFrameRate(int r) 
  {
    seqFrameRate = r;
  }
  void ToonSequence() {
    // for (int i = 0; i < images.length; i++)
    // {
    // images[i] = new PImage(LOOP_WIDTH, LOOP_HEIGHT);
    // }
  }

  void resetMovie() {
    captureFrameNum = 0;
  }

  void deleteFrame() 
  {
    // the former frame goes to the garbage collector
    if (captureFrameNum > 0) 
    {
      captureFrameNum--;
    }
  }

  void addFrame()
  {
    if (captureFrameNum < LOOP_MAX_NUM_FRAME) 
    {
      // We use new here, because it was not possible to overwrite an existing image.
      images[captureFrameNum] = new PImage(LOOP_WIDTH, LOOP_HEIGHT);
      images[captureFrameNum].copy(cam, 0, 0, LOOP_WIDTH, LOOP_HEIGHT, 0, 0, LOOP_WIDTH, LOOP_HEIGHT);

      captureFrameNum++;
      oscSendEditFrameNum(captureFrameNum);
      isFlashing = 1;
    }
    else {
      println("Reached max number of frames : " + LOOP_MAX_NUM_FRAME);
    }
  }
  void loopFrame() 
  {
    if (playFrameNum < captureFrameNum - 1) 
    {
      playFrameNum++;
    } else {
      playFrameNum = 0;
    }
  }
}

/**
One thread to save a movie
 */
class SaveThread extends Thread 
{
  PApplet theParent;
  int seqNum;
  /**
   * Args: PApplet "this", sequence number
   */
  SaveThread(PApplet p, int sequenceNum) 
  {
    seqNum = sequenceNum;
    theParent = p;
    println("saving " + seqNum);
    }

  public void run() {
      // Create MovieMaker object with size, filename,
      // compression codec and quality, framerate
      String file_name;
      if (sequences[seqNum].captureFrameNum == 0) {
          println("no frame to save!");
      } else {
          file_name = "toonloop_" + String.valueOf(year()) + "_"
                  + String.valueOf(month()) + "_" + String.valueOf(day())
                  + "_" + String.valueOf(hour()) + "h"
                  + String.valueOf(minute()) + "m"
                  + String.valueOf(second()) + ".mov";
          println("saving to " + file_name + " :");

          // enable saved message.
          is_displaying_saved_message = SAVED_MESSAGE_DURATION;
          saved_file_name = file_name;

          // movieOut = new MovieMaker(theParent, LOOP_WIDTH, LOOP_HEIGHT,
          // file_name, FRAME_RATE, LOOP_CODEC, LOOP_QUALITY);
          for (int i = 0; i < sequences[seqNum].captureFrameNum; i++) {
              print(i + " ");
              // movieOut.addFrame(sequences[seqNum].images[i].pixels,
              // LOOP_WIDTH, LOOP_HEIGHT);
          }
          println();
          // movieOut.finish();
          println("done saving");
          // movieOut = null;
      }
  }
 }
